import 'dart:async';

import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'lyrics.dart';
import 'lyrics_controller.dart';
import 'lyrics_painter.dart';

class LyricWidget extends StatefulWidget {
  final List<Lyrics> lyrics;
  final List<Lyrics> remarkLyrics;
  final Size size;
  final double currentProgress;
  final LyricController controller;
  TextStyle lyricStyle;
  TextAlign lyricAlign;
  TextDirection textDirection;
  TextStyle remarkStyle;
  TextStyle currLyricStyle;
  TextStyle currRemarkLyricStyle;
  TextStyle draggingLyricStyle;
  TextStyle draggingRemarkLyricStyle;
  final double lyricGap;
  final double remarkLyricGap;
  bool enableDrag;

  //Lyric brush array
  List<TextPainter> lyricTextPaints = [];

  //Translation/transliteration lyrics brush array
  List<TextPainter> subLyricTextPaints = [];

  //Maximum font width
  double lyricMaxWidth;

  LyricWidget(
      {Key key,
      @required this.lyrics,
      this.remarkLyrics,
      @required this.size,
      this.currentProgress,
      this.controller,
      this.lyricStyle,
      this.lyricAlign,
      this.textDirection,
      this.remarkStyle,
      this.currLyricStyle,
      this.lyricGap: 10,
      this.remarkLyricGap: 20,
      this.draggingLyricStyle,
      this.draggingRemarkLyricStyle,
      this.enableDrag: true,
      this.lyricMaxWidth,
      this.currRemarkLyricStyle})
      : assert(enableDrag != null),
        assert(lyrics != null && lyrics.isNotEmpty),
        assert(size != null),
        assert(controller != null) {
    this.lyricStyle ??= TextStyle(
      color: Colors.grey,
      fontSize: 14,
    );
    this.lyricAlign ??= TextAlign.center;
    this.remarkStyle ??= TextStyle(color: Colors.black, fontSize: 14);
    this.currLyricStyle ??= TextStyle(color: Colors.red, fontSize: 14);
    this.currRemarkLyricStyle ??= this.currLyricStyle;
    this.draggingLyricStyle ??= lyricStyle.copyWith(color: Colors.greenAccent);
    this.draggingRemarkLyricStyle ??=
        remarkStyle.copyWith(color: Colors.greenAccent);

    //Lyrics to brush
    lyricTextPaints.addAll(lyrics
        .map(
          (l) => TextPainter(
              text: TextSpan(text: l.lyrics, style: lyricStyle),
              textDirection: textDirection,
              textAlign: lyricAlign),
        )
        .toList());

    //Translation/Transliteration lyrics to brush
    if (remarkLyrics != null && remarkLyrics.isNotEmpty) {
      subLyricTextPaints.addAll(remarkLyrics
          .map((l) => TextPainter(
              text: TextSpan(text: l.lyrics, style: remarkStyle),
              textDirection: TextDirection.ltr,
              textAlign: lyricAlign))
          .toList());
    }
  }

  @override
  _LyricWidgetState createState() => _LyricWidgetState();
}

class _LyricWidgetState extends State<LyricWidget> {
  LyricPainter _lyricPainter;
  double totalHeight = 0;

  @override
  void initState() {
    widget.controller.draggingComplete = () {
      cancelTimer();
      widget.controller.progress = widget.controller.draggingProgress;
      _lyricPainter.draggingLine = null;
      widget.controller.isDragging = false;
    };
    WidgetsBinding.instance.addPostFrameCallback((call) {
      totalHeight = computeScrollY(widget.lyrics.length - 1);
    });
    widget.controller.addListener(() {
      var curLine =
          findLyricIndexByDuration(widget.controller.progress, widget.lyrics);
      if (widget.controller.oldLine != curLine) {
        _lyricPainter.currentLyricIndex = curLine;
        if (!widget.controller.isDragging) {
          if (widget.controller.vsync == null) {
            _lyricPainter.offset = -computeScrollY(curLine);
          } else {
            animationScrollY(curLine, widget.controller.vsync);
          }
        }
        widget.controller.oldLine = curLine;
      }
    });
    super.initState();
  }

  //Because the height of the blank line is inconsistent with the height of the non-blank line, get the position of the non-blank line
  int getNotEmptyLineHeight(List<Lyrics> lyrics) => lyrics.indexOf(
      lyrics.firstWhere((lyric) => lyric.lyrics.trim().isNotEmpty,
          orElse: () => lyrics.first));

  @override
  Widget build(BuildContext context) {
    if (widget.lyricMaxWidth == null ||
        widget.lyricMaxWidth == double.infinity) {
      widget.lyricMaxWidth = MediaQuery.of(context).size.width;
    }

    _lyricPainter = LyricPainter(
        widget.lyrics, widget.lyricTextPaints, widget.subLyricTextPaints,
        vsync: widget.controller.vsync,
        subLyrics: widget.remarkLyrics,
        lyricTextStyle: widget.lyricStyle,
        subLyricTextStyle: widget.remarkStyle,
        currLyricTextStyle: widget.currLyricStyle,
        lyricGapValue: widget.lyricGap,
        lyricMaxWidth: widget.lyricMaxWidth,
        subLyricGapValue: widget.remarkLyricGap,
        draggingLyricTextStyle: widget.draggingLyricStyle,
        draggingSubLyricTextStyle: widget.draggingRemarkLyricStyle,
        currSubLyricTextStyle: widget.currRemarkLyricStyle);
    _lyricPainter.currentLyricIndex =
        findLyricIndexByDuration(widget.controller.progress, widget.lyrics);
    if (widget.controller.isDragging) {
      _lyricPainter.draggingLine = widget.controller.draggingLine;
      _lyricPainter.offset = widget.controller.draggingOffset;
    } else {
      _lyricPainter.offset = -computeScrollY(_lyricPainter.currentLyricIndex);
    }
    return widget.enableDrag
        ? GestureDetector(
            onVerticalDragUpdate: (e) {
              cancelTimer();
              double temOffset = (_lyricPainter.offset + e.delta.dy);
              if (temOffset < 0 && temOffset >= -totalHeight) {
                widget.controller.draggingOffset = temOffset;
                widget.controller.draggingLine =
                    getCurrentDraggingLine(temOffset + widget.lyricGap);
                _lyricPainter.draggingLine = widget.controller.draggingLine;
                widget.controller.draggingProgress =
                    widget.lyrics[widget.controller.draggingLine].startTime +
                        Duration(milliseconds: 1);
                widget.controller.isDragging = true;
                _lyricPainter.offset = temOffset;
              }
            },
            onVerticalDragEnd: (e) {
              cancelTimer();
              widget.controller.draggingTimer = Timer(
                  widget.controller.draggingTimerDuration ??
                      Duration(seconds: 3), () {
                resetDragging();
              });
            },
            child: buildCustomPaint(),
          )
        : buildCustomPaint();
  }

  CustomPaint buildCustomPaint() {
    return CustomPaint(
      painter: _lyricPainter,
      size: widget.size,
    );
  }

  void resetDragging() {
    _lyricPainter.currentLyricIndex =
        findLyricIndexByDuration(widget.controller.progress, widget.lyrics);

    widget.controller.previousRowOffset = -widget.controller.draggingOffset;
    animationScrollY(_lyricPainter.currentLyricIndex, widget.controller.vsync);
    _lyricPainter.draggingLine = null;
    widget.controller.isDragging = false;
  }

  int getCurrentDraggingLine(double offset) {
    for (int i = 0; i < widget.lyrics.length; i++) {
      var scrollY = computeScrollY(i);
      if (offset > -1) {
        offset = 0;
      }
      if (offset >= -scrollY) {
        return i;
      }
    }
    return widget.lyrics.length;
  }

  void cancelTimer() {
    if (widget.controller.draggingTimer != null) {
      if (widget.controller.draggingTimer.isActive) {
        widget.controller.draggingTimer.cancel();
        widget.controller.draggingTimer = null;
      }
    }
  }

  animationScrollY(currentLyricIndex, TickerProvider tickerProvider) {
    var animationController = widget.controller.animationController;
    if (animationController != null) {
      animationController.stop();
    }
    animationController = AnimationController(
      vsync: tickerProvider,
      duration: Duration(milliseconds: 300),
    )..addStatusListener((status) {
        if (status == AnimationStatus.completed) {
          animationController.dispose();
          animationController = null;
        }
      });
    // Calculate the current row offset
    var currentRowOffset = computeScrollY(currentLyricIndex);
    //Do not perform animation if the offsets are the same
    if (currentRowOffset == widget.controller.previousRowOffset) {
      return;
    }
    // The start is the previous line and the end point is the current line
    Animation animation = Tween<double>(
            begin: widget.controller.previousRowOffset, end: currentRowOffset)
        .animate(animationController);
    widget.controller.previousRowOffset = currentRowOffset;
    animationController.addListener(() {
      _lyricPainter.offset = -animation.value;
    });
    animationController.forward();
  }

  //Get the location of the lyrics according to the current duration
  int findLyricIndexByDuration(Duration curDuration, List<Lyrics> lyrics) {
    for (int i = 0; i < lyrics.length; i++) {
      if (curDuration >= lyrics[i].startTime &&
          curDuration <= lyrics[i].endTime) {
        return i;
      }
    }
    return 0;
  }

  // Calculate the offset between the incoming line and the first line
  double computeScrollY(int curLine) {
    double totalHeight = 0;
    for (var i = 0; i < curLine; i++) {
      var currPaint = widget.lyricTextPaints[i]
        ..text = TextSpan(
          text: widget.lyrics[i].lyrics,
          style: widget.lyricStyle,
        );
      currPaint.layout(maxWidth: widget.lyricMaxWidth);
      totalHeight += currPaint.height + widget.lyricGap;
    }
    if (widget.remarkLyrics != null) {
      //Increase the offset of the translated lyrics before the current line
      widget.remarkLyrics
          .where(
              (subLyric) => subLyric.endTime <= widget.lyrics[curLine].endTime)
          .toList()
          .forEach((subLyric) {
        var currentPaint = widget
            .subLyricTextPaints[widget.remarkLyrics.indexOf(subLyric)]
          ..text = TextSpan(text: subLyric.lyrics, style: widget.remarkStyle);
        currentPaint.layout(maxWidth: widget.lyricMaxWidth);
        ;
        totalHeight += widget.remarkLyricGap + currentPaint.height;
      });
    }
    return totalHeight;
  }
}
